<?php

namespace Schaffen\Transaction\Transaction\Services;

use Illuminate\Support\Facades\DB;
use Schaffen\Core\Company\Models\Branch;
use Schaffen\Core\Company\Models\CostCenter;
use Schaffen\Core\Employee\Models\Employee;
use Schaffen\Core\Person\Models\Person;
use Schaffen\Core\Product\Models\ProductVariety;
use Schaffen\Core\Utils\Managers\Utilitarian;
use Schaffen\Transaction\Transaction\Interfaces\IGenerateFinance;
use Schaffen\Transaction\Transaction\Interfaces\IGenerateInvoice;
use Schaffen\Transaction\Transaction\Interfaces\IInformPaymentForm;
use Schaffen\Transaction\Transaction\Interfaces\IManipulateWarehouseInventory;
use Schaffen\Transaction\Transaction\Interfaces\ITransaction;
use Schaffen\Transaction\Transaction\Interfaces\IWithCommission;
use Schaffen\Transaction\Transaction\Models\Transaction;
use Schaffen\Transaction\Transaction\Models\TransactionAssociate;
use Schaffen\Transaction\Transaction\Models\TransactionItem;
use Schaffen\Transaction\Transaction\Models\TransactionItemDelivery;
use Schaffen\Transaction\Transaction\Models\TransactionSplit;
use Schaffen\WarehouseInventory\WarehouseInventory\Models\WarehouseInventory;

class TransactionGenerate implements ITransaction
{

   private IGenerateFinance $generate_finance;
   private IInformPaymentForm $payment_form;
   private IManipulateWarehouseInventory $warehouse_manipulate;
   private IGenerateInvoice $invoice;
   private IWithCommission $comission_item;
   private $transaction;
   private $util;
   private array $transaction_itens = [];
   private array $warehouse_movement = [];
   private array $itens = [];
   private array $payment_forms = [];
   private array $finance = [];
   private array $input = [];
   private array $commissions = [];
   private array $transfer = [];


   public function __construct(
      IGenerateFinance $generate_finance,
      IInformPaymentForm $payment_form,
      IManipulateWarehouseInventory $warehouse_manipulate,
      IGenerateInvoice $invoice,
      IWithCommission $comission_item
   )
   {
      $this->generate_finance = $generate_finance;
      $this->payment_form = $payment_form;
      $this->warehouse_manipulate = $warehouse_manipulate;
      $this->invoice = $invoice;
      $this->comission_item = $comission_item;
      $this->util = new Utilitarian();
   }

   public function realizarSetup(array $input = [])
   {

      if (isset($input['itens'])) {
         $this->itens = $input['itens'];
         unset($input['itens']);
      }
      if (isset($input['avista_aprazo'])) {
         $input['trn_tipo_pagamento'] = $input['avista_aprazo'];
         unset($input['avista_aprazo']);
      }
      if (isset($input['comissoes'])) {
         $this->commissions = $input['comissoes'];
         unset($input['comissoes']);
      }

      if (isset($input['financeiro'])) {
         $this->finance = $input['financeiro'];
         unset($input['financeiro']);
      }

      if (isset($input['formas_pagamento'])) {
         $this->payment_forms = $input['formas_pagamento'];
         unset($input['formas_pagamento']);
      }

      if (empty($input['trn_hash'])) {
         $input['trn_hash'] = sha1(md5(time())); //. Auth::user()->id;
      }
      if (isset($input['transferencia'])) {
         $this->transfer = $input['transferencia'];
         unset($input['transferencia']);
      }

      if (isset($input['amb_codigo'])) { //código do ambiente para caso o cliente esteja utilizando o módulo de caixa
         $this->amb_codigo = $input['amb_codigo'];
         unset($input['amb_codigo']);
      }

      $input['trn_tipo'] = strtoupper($input['trn_tipo']);

      $this->input = $input;
      unset($input['cliente']);

      $this->transaction = Transaction::firstOrNew($input);

      return $this->transaction;

   }

   public function processarTransacao()
   {
      if (!empty($this->input['cliente']['trn_cliente_fk_pes_codigo'])) {
         $this->transaction->cliente()->associate(Person::findOrFail($this->input['cliente']['trn_cliente_fk_pes_codigo']));
      }

      $this->transaction->filial()->associate(Branch::findOrFail($this->input['trn_fk_fil_codigo']));
      $this->transaction->centro_custo()->associate(CostCenter::findOrFail($this->finance[0]['ccu']));
   }

   public function processarItens()
   {
      $i = 0;
      foreach ($this->itens as $item) {
         $this->transaction_itens[$i]['entregas'] = [];
         $this->transaction_itens[$i]['desconto'] = [];
         $agenda_entregas = isset($item['agenda']) ? $item['agenda'] : [];
         $descontos = isset($item['desconto']) ? $item['desconto'] : [];
         unset($item['agenda']);
         unset($item['desconto']);

         $transacao_item = new TransactionItem($item);
         $transacao_item->product_variety_pack()->associate(ProductVariety::findOrFail($item["trit_fk_pve_codigo"]));
         $transacao_item->responsavel()->associate(Person::findOrFail($item["trit_responsavel_fk_pes_codigo"]));
         $this->transaction_itens[$i]['objeto'] = $transacao_item;

         foreach ($agenda_entregas as $entrega) {
            $entrega_objeto = new TransactionItemDelivery($entrega);
            if (isset($entrega["trie_entregador_fk_pes_codigo"]) && $entrega["trie_entregador_fk_pes_codigo"] != null) {
               $entrega_objeto->entregador()->associate(Person::findOrFail($entrega["trie_entregador_fk_pes_codigo"]));
            }

            $this->transaction_itens[$i]['entregas'][] = $entrega_objeto;
         }

         if ($this->transaction['trn_tipo'] == 'TRANSFERENCIA') {
            $this->transaction_itens[$i]['tri_codigo'] = $item['tri_codigo'];
         }

         $i++;
      }


   }

   public function getTransaction()
   {
      return $this->transaction;
   }

   public function salvarTransacao()
   {

      try {
         DB::beginTransaction();
         $this->transaction->save();

         if ($this->transaction['trn_tipo'] == 'TRANSFERENCIA') {
            $transacao_associacao = new TransactionAssociate(
               [
                  "tras_type" => 'transferencia',
                  "tras_id" => $this->transfer['tra_codigo']
               ]);
            $transacao_associacao->transacao()->associate($this->transaction);
            $transacao_associacao->save();
         }

         foreach ($this->transaction_itens as $item) {
            $item['objeto']->trasacao()->associate($this->transaction);
            $item['objeto']->save();
            $item['objeto']->desconto()->saveMany($item['desconto']);
            $item['objeto']->entregas()->saveMany($item['entregas']);
         }

         $this->movimentarEstoque();
         $this->atribuirFormaPagamento();
         $this->gerarMovimentacaoFinanceira();
         $this->gerarComissao();
         $this->gerarNota();

         DB::commit();
      } catch (\Exception $ex) {
         DB::rollback();
         throw $ex;
      }
      return $this->transaction;
   }

   public function movimentarEstoque()
   {
      foreach ($this->transaction_itens as $item) {
         foreach ($item['entregas'] as $entrega) {
            if (!is_null($entrega->trie_data_efetivada) && !$entrega->trie_processado) {
               $item_entrega = [
                  'quantidade' => $entrega->trie_quantidade,
                  'valor_unitario' => $item['objeto']->trit_valor_unitario,
                  'serial' => $item['objeto']->trit_serial,
                  'embalagem' => $item['objeto']->trit_fk_pve_codigo,
               ];

               $loe_origem = WarehouseInventory::where('loe_tipo', 'FUNCIONARIO')->where('loe_fk_pes_codigo', $entrega->trie_entregador_fk_pes_codigo)->first();

               if (empty($loe_origem)) {
                  throw new \Exception('Não há itens suficientes no estoque do vendedor.');
               }

               $this->warehouse_manipulate->setup($loe_origem->loe_codigo);

               $this->warehouse_movement['credito'][] = $this->warehouse_manipulate->creditarEstoque($item_entrega, $this->transaction);
               $entrega->trie_processado = 1;
               $entrega->save();
            }
         }
      }

   }

   //verifica os 3 tipos de pagamento e cria as instancias a partir disso
   public function atribuirFormaPagamento()
   {
      if (is_array($this->payment_forms)) {
         foreach ($this->payment_forms as $forma_pagamento) {
            $this->payment_form->geraEntrada($forma_pagamento, $this->transaction);
         }
      }

   }

   public function gerarMovimentacaoFinanceira($modelo = '')
   {
      bcscale(2);
      $valor_total = 0;
      $formas_pagamento = $this->transaction->formas_pagamento ?? [];
      foreach ($formas_pagamento as $fpg) {
         if ($fpg->tfp_fk_fop_codigo == -1) {
            continue;
         }

         $parcelas = TransactionSplit::where('trp_fk_tfp_codigo', $fpg->tfp_codigo)->get();

         $transacao_itens = ($this->transaction->itens);
         $agrupado = $transacao_itens->groupBy('trit_responsavel_fk_pes_codigo')->toArray();
         $rateios_vendedores = [];
         foreach ($agrupado as $vendedor_chave => $itens) {
            $rateios_vendedores[$vendedor_chave]['pcg'] = Employee::where('fun_fk_pes_codigo', $vendedor_chave)->first()->fun_fk_pcg_codigo_caixa;
            if (isset($this->finance)) {
               foreach ($this->finance as $financeiro) {
                  if (isset($financeiro['ccu'])) {
                     $rateios_vendedores[$vendedor_chave]['ccu'] = $financeiro['ccu'];
                  };
                  if (isset($financeiro['pcg'])) {
                     $rateios_vendedores[$vendedor_chave]['pcg'] = $financeiro['pcg'];
                  };
               }

            }

            foreach ($itens as $item) {
               if (!isset($rateios_vendedores[$vendedor_chave]['valor'])) {
                  $rateios_vendedores[$vendedor_chave]['valor'] = 0;
               }

               $rateios_vendedores[$vendedor_chave]['valor'] = bcadd($rateios_vendedores[$vendedor_chave]['valor'], bcmul($item['trit_quantidade'], $item['trit_valor_unitario']));
               $valor_total = bcadd($valor_total, $rateios_vendedores[$vendedor_chave]['valor']);
            }
         }
         $pessoa = !empty($this->input['cliente']['trn_cliente_fk_pes_codigo']) ? $this->input['cliente']['trn_cliente_fk_pes_codigo'] : -999;
         $filial = $this->input['trn_fk_fil_codigo'];
         $observacao = isset($this->input['trn_obervacao']) ? $this->input['trn_obervacao'] : '';
         $descricao = "MVF referente a " . $this->transaction->trn_tipo . " #" . $this->transaction->trn_codigo;
         $valor = $fpg->tfp_valor;
         $valor_total = collect($rateios_vendedores)->sum('valor');

         $rateios_proporcionais = $this->util->valorRateadoProporcional($valor_total, $valor, $rateios_vendedores);


         $data_inicial = $fpg->tfp_primeiro_vencimento;
         $vencimento_sempre = $fpg->tfp_vencimento_sempre;
         $qtd_dias = $fpg->tfp_dias_entre_parcelas;
         $fop_codigo = $fpg->tfp_fk_fop_codigo;

         $input = compact(
            'pessoa',
            'filial',
            'observacao',
            'descricao',
            'valor',
            'data_inicial',
            'vencimento_sempre',
            'qtd_dias',
            'fop_codigo'
         );


         $mvf = $this->generate_finance->processarMvf(
            $input,
            $rateios_proporcionais,
            $fpg->tfp_numero_vezes,
            $parcelas
         );
//
//         if (is_array($mvf)) {
//            foreach ($mvf['mvfs'] as $mvf) {
//               if ($mvf instanceof MovimentacaoFinanceira) {
//                  $transacao_associacao = new TransacaoAssociacao(
//                     [
//                        "tras_type" => 'mvf',
//                        "tras_id" => $mvf->mvf_codigo
//                     ]);
//                  $transacao_associacao->transacao()->associate($this->transacao);
//                  $transacao_associacao->save();
//               }
//            }
//         }
      }
   }

   public function gerarComissao()
   {
//      if (!empty($this->comissoes)) {
//         $this->tem_comissao = new ComissionaValorVendaPorVendedor();
//      }
//      $this->tem_comissao->comissionar($this->transacao, $this->transacao_itens, $this->comissoes);
   }

   public function gerarNota()
   {

      $nota_fiscal = $this->invoice->processarNotaFiscal($this->transaction, $this->transaction_itens);

//      if ($nota_fiscal instanceof NotaFiscal) {
//         $transacao_associacao = new TransacaoAssociacao(
//            [
//               "tras_type" => 'nota-fiscal',
//               "tras_id" => $nota_fiscal->nof_codigo
//            ]);
//         $transacao_associacao->transacao()->associate($this->transacao);
//         $transacao_associacao->save();
//      }

   }


}
